#include <stdio.h>
#include <stdlib.h>
#include "LRSolver.h"


void LRSolver::AddContractedSubproblem(Problem *P, int n, double x_cost, int N, int* mapping_to_parent, int* mapping_to_child)
{
	if (!S_being_contracted) { printf("incorrect call to AddContractedSubproblem()\n"); exit(1); }

	Subproblem* S = _Slist.New(2);
	S->Init(this, P, n, S_being_contracted->options);

	S->parent = S_being_contracted;

	*S_being_contracted_list_last = S;
	S_being_contracted_list_last = &S->next;
	S->next = NULL;

	S->xi_cost = x_cost;

	S->N = N;
	S->mapping_to_parent = mapping_to_parent;
	S->mapping_to_child = mapping_to_child;

	S->P->parent_subproblem_id = S_being_contracted->P->parent_subproblem_id;
	S->P->child_subproblem_id = child_subproblem_num ++;
	if (options.verbose >= 3)
	{
		if (child_subproblem_num > 1) printf(" +");
		printf(" %d", S->n);
	}
}


bool LRSolver::Subproblem::Contract(Problem::Atom a0)
{
	if (options->contract == 0 && type == 1) return false;

	if (LR->S_being_contracted || type >= 2)
	{
		printf("error in ContractSubproblem()\n"); exit(1);
	}
	LR->S_being_contracted = this;
	LR->S_being_contracted_list_last = &next;
	LR->child_subproblem_num = 0;

	int t, Tnum;
	Subproblem* T;
	for (T=next, Tnum=0; T; T=T->next, Tnum++) {}

	Subproblem** Told = new Subproblem*[Tnum > 0 ? Tnum : 1];
	for (T=next, Tnum=0; T; T=T->next, Tnum++) Told[Tnum] = T;

	if (next)
	{
		delete [] next->mapping;
		delete [] next->xi;
		next = NULL;
	}

	bool delete_a0 = false;
	if (!a0)
	{
		double* lambda = LR->D.GetLambda(this);
		a0 = P->min_oracle(lambda);
		delete_a0 = true;
	}

	if (LR->options.verbose >= 3)
	{
		printf("contract "); P->print_id(); printf(": %d => ", n);
	}

	double* xi_average = NULL;
	if (options->contract == 2)
	{
		xi_average = LR->D.GetBuf();
		int k;
		for (k=0; k<n; k++)
		{
			int i = mapping[k];
			xi_average[k] = (LR->nu[i] - mui[k] / LR->var_counts[i]) / LR->GAMMA;
		}
	}

	bool success = false;
	if (options->contract)
	{
		success = P->contract(xi, xi_cost, a0, xi_average);
	}

	int old_type = type;
	int new_type = (success) ? 0 : 1;
	if (old_type != new_type)
	{
		LR->_Slist.ChangeType(this, new_type);
	}

	if (LR->options.verbose >= 3)
	{
		printf("\n");
	}

	if (next)
	{
		int n_new = 0, N_new = 0;
		for (T=next; T; T=T->next) { n_new += T->n; N_new += T->N; }

		int* mapping_new = new int[N_new];
		double* xi_new = new double[2*n_new];
		double* mui_new = xi_new + n_new;

		n_new = N_new = 0;
		for (T=next; T; T=T->next)
		{
			T->mapping = mapping_new + N_new; 
			T->xi = xi_new + n_new;
			T->mui = mui_new + n_new;
			n_new += T->n;
			N_new += T->N;

			XMAPPING_FOR(T, r, k, i,\
			{\
				T->mui[k] = 0;\
			},\
			{\
				int p = T->mapping_to_parent[r];\
				T->mapping[r] = mapping[p];\
				T->xi[k] = xi[p];\
				T->mui[k] += mui[p];\
			})

			T->SetDotProductWeights();
		}
	}

	if (old_type == 1 && new_type == 0)
	{
		if (options->conv)
		{
			TransformContractedAtomsConv10(a0);
		}
		else
		{
			TransformContractedAtomsCache10(a0);
		}
	}
	else if (old_type == 0)
	{
		if (options->conv)
		{
			TransformContractedAtomsConv(Tnum, Told, a0, (new_type == 1) ? this : next);
		}
		else
		{
			TransformContractedAtomsCache(Tnum, Told, a0, (new_type == 1) ? this : next);
		}
	}

	if (delete_a0) P->destroy_atom(a0);
	for (t=0; t<Tnum; t++) Told[t]->Delete();
	delete [] Told;

	LR->S_being_contracted = NULL;

	return success;
}

void LRSolver::Subproblem::TransformContractedAtomsConv10(Problem::Atom a0)
{
	assert(atom_num >= 1);

	int k;
	Subproblem* R;

	for (k=0; k<atom_num; k++)
	{
		double weight = atoms[k].weight;
		if (weight > 0)
		{
			P->set_current_atom(atoms[k].atom);
			for (R=next; R; R=R->next)
			{
				Problem::Atom a = R->P->get_current_atom();
				if (a)
				{
					int p = R->AddAtom(a, R->P->compute_atom_cost(a));
					R->atoms[p].weight += weight;
				}
			}
		}
	}
	P->set_current_atom(a0); 
	for (R=next; R; R=R->next)
	{
		Problem::Atom a = R->P->get_current_atom();
		if (!a)
		{
			printf("incorrect contraction?\n"); exit(1); 
		}
		R->AddAtom(a, R->P->compute_atom_cost(a));

		R->Recompute_x(true, true);
	}
	for (k=atom_num-1; k>=0; k--)
	{
		RemoveAtom(k);
	}
}

void LRSolver::Subproblem::TransformContractedAtomsConv(int Tnum, Subproblem** Told, Problem::Atom a0, Subproblem* Rfirst)
{
	assert(options->conv);

	int t;
	Subproblem* T;
	Subproblem* R;

	double* sum = new double[Tnum > 0 ? Tnum : 1];
	int* ind = new int[Tnum > 0 ? Tnum : 1];

	for (t=0; t<Tnum; t++)
	{
		T = Told[t];
		assert(T->atom_num >= 1);
		ind[t] = 1;
		sum[t] = T->atoms[0].weight;
		T->P->set_current_atom(T->atoms[0].atom);
	}
	double sum_prev = 0;
	while ( 1 )
	{
		int t_min = -1;
		for (t=0; t<Tnum; t++)
		{
			if (ind[t]<Told[t]->atom_num && (t_min<0 || sum[t_min] > sum[t])) t_min = t;
		}

		double weight;
		if (t_min >= 0) { weight = sum[t_min] - sum_prev; sum_prev = sum[t_min]; }
		else weight = 1 - sum_prev;
		if (weight > 0)
		{
			for (R=Rfirst; R; R=R->next)
			{
				Problem::Atom a = R->P->get_current_atom();

				if (!a) continue;
				int p = R->AddAtom(a, R->P->compute_atom_cost(a));
				R->atoms[p].weight += weight;
			}
		}

		if (t_min < 0) break;

		T = Told[t_min];
		T->P->set_current_atom(T->atoms[ind[t_min]].atom);
		if (ind[t_min] < T->atom_num) sum[t_min] += T->atoms[ind[t_min]].weight;
		ind[t_min] ++;
	}

	delete [] sum;
	delete [] ind;

	P->set_current_atom(a0);
	for (R=Rfirst; R; R=R->next)
	{
		Problem::Atom a = R->P->get_current_atom();
		if (!a)
		{
			printf("incorrect contraction?\n"); exit(1); 
		}
		R->AddAtom(a, R->P->compute_atom_cost(a));

		R->Recompute_x(true, true);
	}
}


void LRSolver::Subproblem::TransformContractedAtomsCache10(Problem::Atom a0)
{
	assert(!options->conv);

	int a;
	Subproblem* R;

	for (a=0; a<atom_num; a++)
	{
		P->set_current_atom(atoms[a].atom);
		for (R=next; R; R=R->next)
		{
			Problem::Atom a = R->P->get_current_atom();
			if (a)
			{
				int p = R->AddAtom(a, R->P->compute_atom_cost(a));
				R->TouchAtom(p);
			}
		}
	}
	P->set_current_atom(a0);
	for (R=next; R; R=R->next)
	{
		R->PruneAtoms();
		Problem::Atom a = R->P->get_current_atom();
		if (a)
		{
			int p = R->AddAtom(a, R->P->compute_atom_cost(a));
			R->TouchAtom(p);
		}
	}
}


void LRSolver::Subproblem::TransformContractedAtomsCache(int Tnum, Subproblem** Told, Problem::Atom a0, Subproblem* Rfirst)
{
	assert(!options->conv);

	int t, a, atom_num_max = 0;
	Subproblem* T;
	Subproblem* R;

	for (t=0; t<Tnum; t++)
	{
		T = Told[t];
		if (atom_num_max < T->atom_num) atom_num_max = T->atom_num;
	}

	P->set_current_atom(a0);
	for (a=0; a<atom_num_max; a++)
	{
		for (t=0; t<Tnum; t++)
		{
			T = Told[t];
			if (T->atom_num > 0) T->P->set_current_atom(T->atoms[a % T->atom_num].atom);
		}
		for (R=Rfirst; R; R=R->next)
		{
			Problem::Atom a = R->P->get_current_atom();
			if (a)
			{
				int p = R->AddAtom(a, R->P->compute_atom_cost(a));
				R->TouchAtom(p);
			}
		}
	}
	P->set_current_atom(a0);
	for (R=Rfirst; R; R=R->next)
	{
		Problem::Atom a = R->P->get_current_atom();
		if (!a) { printf("incorrect contraction?\n"); exit(1); }
		R->PruneAtoms();
		int p = R->AddAtom(a, R->P->compute_atom_cost(a));
		R->TouchAtom(p);
	}
}

