#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "options_reader.h"
#include "MRF.h"

#include <map>
#include <random>

MRF* generate_grid_potts(int X, int Y, int _K, int seed, Buffer* buf)
{
	srand(seed);
	rand(); rand();

	int x, y, i, j, e, k;
	int node_num = X*Y;
	int* K = (int*) buf->Alloc(node_num*sizeof(int));
	for (i=0; i<node_num; i++) K[i] = _K;
	MRF* mrf = new MRF(node_num, K, 2*node_num);
	for (x=0; x<X; x++)
	for (y=0; y<Y; y++)
	{
		i = x + y*X;
		for (e=0; e<2; e++)
		{
			if (e == 0) { if (x < X-1) j = i + 1; else continue; }
			else        { if (y < Y-1) j = i + X; else continue; }
			double lambda = rand() % 10;
			if ( 1 )
			{
				PairwiseTermPotts* T = AllocPairwiseTermPotts(K[i], lambda, buf);
				mrf->AddEdge(i, j, T);
			}
			else
			{
				PairwiseTermGeneral* T = AllocPairwiseTermGeneral(K[i], K[j], NULL, buf);
				for (k=0; k<K[i]*K[j]; k++) T->V[k] = lambda;
				for (k=0; k<_K; k++) T->V[k + _K*k] = 0;
				mrf->AddEdge(i, j, T);
			}
		}
	}
	double* D = new double[_K];
	for (i=0; i<node_num; i++)
	{
		for (k=0; k<_K; k++) D[k] = rand() % 10;
		mrf->AddUnary(i, D);
	}
	delete [] D;

	return mrf;
}

MRF* generate_grid(int X, int Y, int Kmin, int Kmax, int seed, Buffer* buf)
{
	srand(seed);
	rand(); rand();

	int x, y, i, j, e, k;
	int node_num = X*Y;
	int* K = (int*) buf->Alloc(node_num*sizeof(int));
	for (i=0; i<node_num; i++) K[i] = Kmin + (rand() % (Kmax-Kmin+1));
	MRF* mrf = new MRF(node_num, K, 2*node_num);
	for (x=0; x<X; x++)
	for (y=0; y<Y; y++)
	{
		i = x + y*X;
		for (e=0; e<2; e++)
		{
			if (e == 0) { if (x < X-1) j = i + 1; else continue; }
			else        { if (y < Y-1) j = i + X; else continue; }
			PairwiseTermGeneral* T = AllocPairwiseTermGeneral(K[i], K[j], NULL, buf);
			for (k=0; k<K[i]*K[j]; k++) T->V[k] = (double)(rand() % 10);
			mrf->AddEdge(i, j, T);
		}
	}

	return mrf;
}

MRF* generate_grid_ising(int X, int Y, int Z, double unary, int seed, Buffer* buf)
{
	srand(seed);
	rand(); rand();
	std::default_random_engine eng;
	eng.seed(rand());
	std::normal_distribution<> dist(0, 1);

	int x, y, z, i, j, e;
	int node_num = X*Y*Z;
	int* K = (int*) buf->Alloc(node_num*sizeof(int));
	for (i=0; i<node_num; i++) K[i] = 2;
	MRF* mrf = new MRF(node_num, K, ((Z>1) ? 3 : 2)*node_num);
	if (unary != 0)
	{
		double D[2];
		D[0] = 0; D[1] = (int)(10*unary);
		for (i=0; i<node_num; i++) mrf->AddUnary(i, D);
	}
	else
	{
		double D[2];
		D[0] = 0; D[1] = 1000;
		x = X/2; y = Y/2; z = Z/2;
		mrf->AddUnary(x + y*X + z*X*Y, D);
	}
	for (x=0; x<X; x++)
	for (y=0; y<Y; y++)
	for (z=0; z<Z; z++)
	{
		i = x + y*X + z*X*Y;
		for (e=0; e<3; e++)
		{
			if      (e == 0) { if (x < X-1) j = i + 1;   else continue; }
			else if (e == 1) { if (y < Y-1) j = i + X;   else continue; }
			else if (e == 2) { if (z < Z-1) j = i + X*Y; else continue; }
			PairwiseTermGeneral* T = AllocPairwiseTermGeneral(2, 2, NULL, buf);
			T->V[0] = T->V[3] = 0;
			T->V[1] = T->V[2] = (int)(10*dist(eng));
			mrf->AddEdge(i, j, T);
		}
	}

	return mrf;
}

MRF* generate_random(int node_num, int edge_num, int Kmin, int Kmax, int seed, Buffer* buf)
{
	srand(seed);
	rand(); rand();

	int i, j, e, k;
	int* K = (int*) buf->Alloc(node_num*sizeof(int));
	for (i=0; i<node_num; i++) K[i] = Kmin + (rand() % (Kmax-Kmin+1));
	MRF* mrf = new MRF(node_num, K, edge_num);
	for (e=0; e<edge_num; e++)
	{
		i = rand()%node_num;
		j = rand()%(node_num-1); if (j>=i) j ++;
		PairwiseTermGeneral* T = AllocPairwiseTermGeneral(K[i], K[j], NULL, buf);
		for (k=0; k<K[i]*K[j]; k++) T->V[k] = (double)(rand() % 10);
		mrf->AddEdge(i, j, T);
	}

	double* D = new double[Kmax];
	for (i=0; i<node_num; i++)
	{
		for (k=0; k<K[i]; k++) D[k] = rand() % 10;
		mrf->AddUnary(i, D);
	}
	delete [] D;

	return mrf;
}


void ReadLROptions(MRF::SRMPOptions& SRMPoptions, MRF::LROptions& LRoptions, OptionsReader& R, int& seed, const char*& loadfile, const char*& savefile)
{
	R.Add(">critical parameter: regularization");
	R.Add("gamma",              LRoptions.global_options.gamma, "gamma");

	R.Add(">termination criteria:");
	R.Add("time_max",             LRoptions.global_options.time_max);
	R.Add("mu_update_max",        LRoptions.global_options.mu_update_max, "max number of updates of 'mu' (i.e. proximal centers)");
	R.Add("mu_eps",               LRoptions.global_options.mu_eps, "stop if ||mu'-mu|| <= mu_eps");
	R.Add(">parameters of PPA (control when and how udate proximal center):");
	R.Add("PPA_flag",             LRoptions.global_options.PPA_flag);
	R.Add("PPA_alpha",            LRoptions.global_options.PPA_alpha);
	R.Add("major_iter#0",       LRoptions.global_options.major_iter_min);
	R.Add("major_iter#1",       LRoptions.global_options.major_iter_max);
	R.Add("major_iter#2",       LRoptions.global_options.major_iter_max_log, "a:b:c = [a, b+c*log_2(1+t)] major iterations in t's proximal step");
	R.Add(">parameters of a major iteration:");
	R.Add("ratio_cache",          LRoptions.global_options.ratio_cache);
	R.Add("ratio_minor",          LRoptions.global_options.ratio_minor);
	R.Add("time_minor_max",       LRoptions.global_options.time_minor_max);
	R.Add("time_cache_max",       LRoptions.global_options.time_cache_max);

	R.Add("minor_iter#0",       LRoptions.global_options.minor_iter_min);
	R.Add("minor_iter#1",       LRoptions.global_options.minor_iter_max, "min:max");
	R.Add("cache_iter#0",       LRoptions.global_options.cache_iter_min);
	R.Add("cache_iter#1",       LRoptions.global_options.cache_iter_max, "min:max");

	R.Add("compute_gap_freq",     LRoptions.global_options.compute_gap_freq);

	const char* ALGS[] = { "FW", "DICG_PFW", "DICG_AFW", "DICG_CACHE" };
	const LRSolver::SubproblemOptions::ALG algs[] = { LRSolver::SubproblemOptions::ALG::FW, LRSolver::SubproblemOptions::ALG::DICG_PFW, LRSolver::SubproblemOptions::ALG::DICG_AFW, LRSolver::SubproblemOptions::ALG::DICG_CACHE };
	int ALG_NUM = sizeof(ALGS) / sizeof (const char*);
	int alg;
	for (alg=0; alg<ALG_NUM; alg++) { if (algs[alg] == LRoptions.subproblem_options.alg) break; }
	R.Add("alg",    alg, ALG_NUM, ALGS);
	LRoptions.subproblem_options.alg = algs[alg];

	R.Add("contract",                LRoptions.subproblem_options.contract);
	R.Add("contract_average_eps",    LRoptions.subproblem_options.contract_average_eps);
	R.Add("contract_threshold",      LRoptions.subproblem_options.contract_threshold);
	R.Add("max_cache_size",          LRoptions.subproblem_options.max_cache_size);
	R.Add("inactive_atom_iter_max",  LRoptions.subproblem_options.inactive_atom_iter_max);
	R.Add("conv",                    LRoptions.subproblem_options.conv);
	R.Add("weight_eps",              LRoptions.subproblem_options.weight_eps, "if weights becomes smaller than weight_eps, then set it to zero");
	R.Add("simplex_accuracy_before", LRoptions.subproblem_options.simplex_accuracy_before);
	R.Add("simplex_accuracy_after",  LRoptions.subproblem_options.simplex_accuracy_after);

	R.Add(">miscellaneous:");
	R.Add("verbose",              LRoptions.global_options.verbose);
	R.Add("loadfile",             loadfile);
	R.Add("savefile",             savefile);

	R.Add("primal_freq", LRoptions.global_options.callback_freq, "frequency of computing primal costs (its time is not counted)");
	R.Add("adjustx_freq", LRoptions.global_options.adjustx_freq, "frequency of adjusting x");
	R.Add("decomposition_method", LRoptions.decomposition_method, "0: first try grid, then OPT + break large trees");
	R.Add(                                                        "1: greedy");
	R.Add(                                                        "2: greedy + break large trees");
	R.Add(                                                        "3: OPT (find decomposition that minimizes # forests - can be slow!)");
	R.Add(                                                        "4: OPT + break large tree");
	R.Add("seed", seed);

	R.Add("");
	R.Add(">optionally, you can initialize LRSolver with SRMP. SRMP options:");
	R.Add("srmp_iter_max",              SRMPoptions.iter_max);
	R.Add("srmp_time_max",              SRMPoptions.time_max);
	R.Add("srmp_eps",                   SRMPoptions.eps);
	R.Add("srmp_compute_solution_freq", SRMPoptions.compute_solution_freq);
	R.Add("srmp_verbose",               SRMPoptions.verbose);

	R.Add("");

	if (!LRoptions.subproblem_options.contract) LRoptions.global_options.minor_iter_max = 1;
	if (!LRoptions.subproblem_options.conv) LRoptions.global_options.adjustx_freq = 5; 

	R.Finish();
}



void Usage(const char* name0)
{
	printf("Usage: %s <instance description> [options]\n", name0);
	printf("  e.g.  example.txt:         read instance from example.txt\n");
	printf("  e.g.  grid5x2#3-4:         2D grid of size 5 x 2, number of labels in [3,4]\n");
	printf("        grid5x2#3-4,seed42:  same, with seed=42\n");
	printf("        grid5x2#potts4:      2D grid of size 5 x 2, potts potentials with 4 labels\n");
	printf("        random5x10#3-4:      random graph with 5 nodes and 10 edges, number of labels in [3,4]\n");
	printf("\nTo see the list of options: %s <instance description> help\n", name0);
	exit(1);
}



MRF* GetMRF(const char* _s, Buffer& buf)
{
	MRF* mrf = NULL;

	int node_num, edge_num, sizeX, sizeY, sizeZ, Kmin, Kmax, seed=42;
	double unary;
	size_t s_len = strlen(_s);
	char* s = new char[s_len+1];
	memcpy(s, _s, s_len+1);

	size_t i;
	const char* save_filename = NULL;

	for (i=s_len-1; i>0; i--)
	{
		if (sscanf(s+i, ",seed%d", &seed) == 1)
		{
			s[i] = 0; 
			s_len = strlen(s);
			break;
		}
	}

	if (sscanf(s, "grid%dx%d#%d-%d", &sizeX, &sizeY, &Kmin, &Kmax) == 4)
	{
		mrf = generate_grid(sizeX, sizeY, Kmin, Kmax, seed, &buf);
	}
	else if (sscanf(s, "grid%dx%d#potts%d", &sizeX, &sizeY, &Kmin) == 3)
	{
		mrf = generate_grid_potts(sizeX, sizeY, Kmin, seed, &buf);
	}
	else if (sscanf(s, "random%d-%d#%d-%d", &node_num, &edge_num, &Kmin, &Kmax) == 4)
	{
		mrf = generate_random(node_num, edge_num, Kmin, Kmax, seed, &buf);
	}
	else if (sscanf(s, "grid-ising%dx%dx%d,%lf", &sizeX, &sizeY, &sizeZ, &unary) == 4)
	{
		mrf = generate_grid_ising(sizeX, sizeY, sizeZ, unary, seed, &buf);
	}
	else if (s_len > 4 && !strcmp(s+s_len-4, ".txt"))
	{
		mrf = new MRF(s, &buf);
	}

	delete [] s;
	
	if (save_filename && mrf)
	{
		mrf->Save(save_filename);
	}

	return mrf;
}

int main(int argc, char* argv[])
{
	Buffer buf(1024);
	
	if (argc < 2) Usage(argv[0]);
	MRF* mrf = GetMRF(argv[1], buf);
	if (!mrf) Usage(argv[0]);
	mrf->PrintStats();
	// mrf->Print();

	int i;
	for (i=1; i<argc-1; i++) argv[i] = argv[i+1];
	argc --;
	OptionsReader R(argc, argv);

	MRF::SRMPOptions SRMPoptions;
	MRF::LROptions LRoptions;
	int seed = 42;
	const char* loadfile = NULL;
	const char* savefile = NULL;
	ReadLROptions(SRMPoptions, LRoptions, R, seed, loadfile, savefile);
	srand(seed);

//	mrf->Print();

	if (SRMPoptions.iter_max > 0 && SRMPoptions.time_max > 0)
	{
		mrf->RunSRMP(SRMPoptions);
	}
	if (LRoptions.global_options.mu_update_max > 0 && LRoptions.global_options.time_max > 0)
	{
		mrf->RunLR(LRoptions);
	}
	if (savefile)
	{
		mrf->Save(savefile);
	}

	delete mrf;
	
	return 0;
}

