#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "MRF.h"

double MRF::ComputeLowerBound()
{
	double LB = 0;

	int dir, k;
	Node* i;
	Edge* e;
	double* A = (double*) rbuffer.Alloc(2*K_max*sizeof(double));
	double* B = A + K_max;

	for (i=nodes; i<nodes+node_num; i++)
	{
		for (k=0; k<i->K; k++) A[k] = i->D[k];
		for (dir=0; dir<2; dir++)
		for (e=i->first[dir]; e; e=e->next[dir])
		{
			for (k=0; k<i->K; k++) A[k] += e->message[dir][k];
		}
		double x_min = A[0];
		for (k=1; k<i->K; k++) { if (x_min > A[k]) x_min = A[k]; }
		LB += x_min;
	}

	for (e=edges; e<edges+edge_num; e++)
	{
		LB += e->T->pass_message_srmp_fw(e->message[0], B, e->i[0]->K, e->i[1]->K);
		B[0] -= e->message[1][0];
		double x_min = B[0];
		for (k=1; k<e->i[1]->K; k++)
		{
			B[k] -= e->message[1][k];
			if (x_min > B[k]) x_min = B[k];
		}
		LB += x_min;
	}
	return LB;
}

void MRF::RunSRMP(SRMPOptions& options)
{
	RemoveSingletons();
	if (node_num == 0)
	{
		RestoreSingletons();
		return;
	}

	InitMessages();
	int iter, k;
	Node* i;
	Edge* e;

	for (e=edges; e<edges+edge_num; e++)
	{
		if (e->i[0] > e->i[1]) e->Swap(&rbuffer);
	}

	double* A = (double*) rbuffer.Alloc(3*K_max*sizeof(double));
	double* B = A + K_max;
	double* C = B + K_max;

	double LB_singletons = 0;

	double time_start = get_time();

	double LB = 0, LB_prev = 0;
	double cost_fw, cost_bw;
	int iter_max = options.iter_max;

	for (iter=0; iter<iter_max; iter++)
	{
		if ((iter > 1 && LB<LB_prev + options.eps) || get_time()-time_start > options.time_max)
			iter_max = iter+1; // last iteration, with compute_solution=true
		LB_prev = LB;

		bool compute_solution = false;
		if (iter != 0 && iter % options.compute_solution_freq == 0) compute_solution = true;
		if (iter == iter_max-1) compute_solution = true;

		// forward pass
		for (i=nodes; i<nodes+node_num; i++)
		{
			memcpy(B, i->D, i->K*sizeof(double));
			int d_in = 0, d_out = 0;
			for (e=i->first[0]; e; e=e->next[0], d_out ++)
			{
				for (k=0; k<i->K; k++) B[k] += e->message[0][k];
			}
			if (compute_solution)
			{
				memcpy(C, B, i->K*sizeof(double));
				for (e=i->first[1]; e; e=e->next[1])
				{
					Node* j = e->i[0];
					for (k=0; k<i->K; k++) C[k] += e->T->get(j->solution, k, j->K, i->K);
				}
				i->solution = 0;
				for (k=1; k<i->K; k++) { if (C[i->solution] > C[k]) i->solution = k; }
			}
			for (e=i->first[1]; e; e=e->next[1], d_in ++)
			{
				Node* j = e->i[0];
				e->T->pass_message_srmp_fw(e->message[0], e->message[1], j->K, i->K);
				for (k=0; k<i->K; k++) B[k] += e->message[1][k];
			}
			int d = (d_in > d_out) ? d_in : d_out;
			for (k=0; k<i->K; k++) B[k] /= d;
			for (e=i->first[0]; e; e=e->next[0])
			{
				for (k=0; k<i->K; k++) e->message[0][k] -= B[k];
			}
		}
		if (compute_solution)
		{
			cost_fw = ComputeCost(NULL);
		}

		// backward pass
		LB = LB_singletons;
		for (i--; i>=nodes; i--)
		{
			memcpy(B, i->D, i->K*sizeof(double));
			int d_in = 0, d_out = 0;
			for (e=i->first[1]; e; e=e->next[1], d_out ++)
			{
				for (k=0; k<i->K; k++) B[k] += e->message[1][k];
			}
			if (compute_solution)
			{
				memcpy(C, B, i->K*sizeof(double));
				for (e=i->first[0]; e; e=e->next[0])
				{
					Node* j = e->i[1];
					for (k=0; k<i->K; k++) C[k] += e->T->get(k, j->solution, i->K, j->K);
				}
				i->solution = 0;
				for (k=1; k<i->K; k++) { if (C[i->solution] > C[k]) i->solution = k; }
			}
			for (e=i->first[0]; e; e=e->next[0], d_in ++)
			{
				Node* j = e->i[1];
				LB += e->T->pass_message_srmp_bw(e->message[0], e->message[1], i->K, e->i[1]->K);
				for (k=0; k<i->K; k++) B[k] += e->message[0][k];
			}
			int d = (d_in > d_out) ? d_in : d_out;
			if (d > 0)
			{
				for (k=0; k<i->K; k++) B[k] /= d;
				if (d_out != d)
				{
					double v = B[0];
					for (k=1; k<i->K; k++) if (v > B[k]) v = B[k];
					LB += v * (d - d_out);
				}
				for (e=i->first[1]; e; e=e->next[1])
				{
					for (k=0; k<i->K; k++) e->message[1][k] -= B[k];
				}
			}
			else
			{
				if (iter == 0)
				{
					double v = i->D[0];
					for (k=1; k<i->K; k++) { if (v > i->D[k]) v = i->D[k]; }
					LB_singletons += v;
					LB += v;
				}
			}
		}
		if (compute_solution)
		{
			cost_bw = ComputeCost(NULL);
		}

		if (options.verbose)
		{
			printf("iter %d (%f secs): LB=%f", iter, get_time() - time_start, LB);
			if (compute_solution) printf(", cost_fw=%f, cost_bw=%f", cost_fw, cost_bw);
			printf("\n");
		}
	}

	RestoreSingletons();
}

void MRF::PrintAllMessages()
{
	Edge* e;
	int dir, k;
	for (e=edges; e<edges+edge_num; e++)
	{
		for (dir=0; dir<2; dir++)
		{
			printf("message %d=>%d: ", (int)(e->i[dir]-nodes), (int)(e->i[dir ^ 1]-nodes));
			for (k=0; k<e->i[dir ^ 1]->K; k++) printf(" %f", e->message[dir ^ 1][k]);
			printf("\n");
		}
	}
}

