#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "opt_transport.h"

void OptTransport::Solve(int A, int B, double* V, double* x, double* y, double* z)
{
	nodeNum = A + B;
	edgeNum = 0;
	edgeNumMax = A*B;
	counter = 0;
	cost = 0;
		
	nodes = (Node*) malloc(nodeNum*sizeof(Node));
	arcs = (Arc*) malloc(2*edgeNumMax*sizeof(Arc));

	memset(nodes, 0, nodeNum*sizeof(Node));
	memset(arcs, 0, 2*edgeNumMax*sizeof(Arc));
	firstActive = &nodes[nodeNum];



	int a, b;
	for (b=0; b<B; b++)
	for (a=0; a<A; a++)
	{
		AddEdge(a, A+b, 1, 0, V[a+A*b]);
	}
	for (a=0; a<A; a++) AddNodeExcess(a, x[a]);
	for (b=0; b<B; b++) AddNodeExcess(A+b, -y[b]);
	SolveMinCost();
	for (a=0; a<A*B; a++)
	{
		z[a] = GetReverseRCap(a);
	}



	free(nodes);
	free(arcs);
}

void OptTransport::Init()
{
	Node* i;
	Arc* a;

	for (a=arcs; a<arcs+2*edgeNum; a++)
	{
		if (a->r_cap > 0 && a->GetRCost() < 0) PushFlow(a, a->r_cap);
	}

	Node** lastActivePtr = &firstActive;
	for (i=nodes; i<nodes+nodeNum; i++)
	{
		if (i->excess > 0)
		{
			*lastActivePtr = i;
			lastActivePtr = &i->next;
		}
		else i->next = NULL;
	}
	*lastActivePtr = &nodes[nodeNum];
}


double OptTransport::Augment(Node* start, Node* end)
{
	double delta = (start->excess < -end->excess) ? start->excess : -end->excess;
	Arc* a;

	for (a=end->parent; a; a=a->sister->head->parent)
	{
		if (delta > a->r_cap) delta = a->r_cap;
	}
	assert(delta > 0);

	end->excess += delta;
	for (a=end->parent; a; a=a->head->parent)
	{
		DecreaseRCap(a, delta);
		a = a->sister;
		IncreaseRCap(a, delta);
	}
	start->excess -= delta;

	return delta;
}

void OptTransport::Dijkstra(Node* start)
{
	assert(start->excess > 0);

	Node* i;
	Node* j;
	Arc* a;
	double d;
	Node* permanentNodes;

	int FLAG0 = ++ counter; // permanently labeled nodes
	int FLAG1 = ++ counter; // temporarily labeled nodes

	start->parent = NULL;
	start->flag = FLAG1;
	queue.Reset();
	queue.Add(start, 0);

	permanentNodes = NULL;

	while ( (i=queue.RemoveMin(d)) )
	{
		if (i->excess < 0)
		{
			double delta = Augment(start, i);
			cost += delta*(d - i->pi + start->pi);
			for (i=permanentNodes; i; i=i->next_permanent) i->pi += d;
			return;
		}

		i->pi -= d;
		i->flag = FLAG0;
		i->next_permanent = permanentNodes;
		permanentNodes = i;

		for (a=i->firstNonsaturated; a; a=a->next)
		{
			j = a->head;
			if (j->flag == FLAG0) continue;
			d = a->GetRCost();
			if (d < 0) d = 0; // to avoid numerical instability
			if (j->flag == FLAG1)
			{
				if (d >= queue.GetKey(j)) continue;
				queue.DecreaseKey(j, d);
			}
			else
			{
				queue.Add(j, d);
				j->flag = FLAG1;
			}
			j->parent = a;
		}
	}
	start->excess = 0;
}


double OptTransport::SolveMinCost()
{
	Node* i;
	//Init();

	while ( 1 )
	{
		i = firstActive;
		if (i == &nodes[nodeNum]) break;
		firstActive = i->next;
		i->next = NULL;
		if (i->excess > 0)
		{
			Dijkstra(i);
			if (i->excess > 0 && !i->next) 
			{ 
				i->next = firstActive; 
				firstActive = i; 
			}
		}
	}

	return cost;
}




inline void OptTransport::AddNodeExcess(NodeId _i, double excess)
{
	assert(_i>=0 && _i<nodeNum);
	nodes[_i].excess += excess;
	if (nodes[_i].excess > 0 && !nodes[_i].next)
	{
		nodes[_i].next = firstActive;
		firstActive = &nodes[_i];
	}
}

inline OptTransport::EdgeId OptTransport::AddEdge(NodeId _i, NodeId _j, double cap, double rev_cap, double cost)
{
	assert(_i>=0 && _i<nodeNum);
	assert(_j>=0 && _j<nodeNum);
	assert(_i!=_j && edgeNum<edgeNumMax);
	assert(cap >= 0);
	assert(rev_cap >= 0);

	Arc *a = &arcs[2*edgeNum];
	Arc *a_rev = a+1;
	edgeNum ++;

	Node* i = nodes + _i;
	Node* j = nodes + _j;

	a -> sister = a_rev;
	a_rev -> sister = a;
	if (cap > 0)
	{
		if (i->firstNonsaturated) i->firstNonsaturated->prev = a;
		a -> next = i -> firstNonsaturated;
		i -> firstNonsaturated = a;
	}
	else
	{
		if (i->firstSaturated) i->firstSaturated->prev = a;
		a -> next = i -> firstSaturated;
		i -> firstSaturated = a;
	}
	a->prev = NULL;
	if (rev_cap > 0)
	{
		if (j->firstNonsaturated) j->firstNonsaturated->prev = a_rev;
		a_rev -> next = j -> firstNonsaturated;
		j -> firstNonsaturated = a_rev;
	}
	else
	{
		if (j->firstSaturated) j->firstSaturated->prev = a_rev;
		a_rev -> next = j -> firstSaturated;
		j -> firstSaturated = a_rev;
	}
	a_rev->prev = NULL;

	a -> head = j;
	a_rev -> head = i;
	a -> r_cap = cap;
	a_rev -> r_cap = rev_cap;
	a -> cost = cost;
	a_rev -> cost = -cost;

	if (a->r_cap > 0 && a->GetRCost() < 0) PushFlow(a, a->r_cap);
	if (a_rev->r_cap > 0 && a_rev->GetRCost() < 0) PushFlow(a_rev, a_rev->r_cap);

	return edgeNum-1;
}

///////////////////////////////////////
///////////////////////////////////////
///////////////////////////////////////

inline void OptTransport::DecreaseRCap(Arc* a, double delta)
{
	a->r_cap -= delta;
	if (a->r_cap == 0)
	{
		Node* i = a->sister->head;
		if (a->next) a->next->prev = a->prev;
		if (a->prev) a->prev->next = a->next;
		else         i->firstNonsaturated = a->next;
		a->next = i->firstSaturated;
		if (a->next) a->next->prev = a;
		a->prev = NULL;
		i->firstSaturated = a;
	}
}

inline void OptTransport::IncreaseRCap(Arc* a, double delta)
{
	if (a->r_cap == 0)
	{
		Node* i = a->sister->head;
		if (a->next) a->next->prev = a->prev;
		if (a->prev) a->prev->next = a->next;
		else         i->firstSaturated = a->next;
		a->next = i->firstNonsaturated;
		if (a->next) a->next->prev = a;
		a->prev = NULL;
		i->firstNonsaturated = a;
	}
	a->r_cap += delta;
}

inline double OptTransport::GetRCap(EdgeId e)
{
	Arc* a = &arcs[2*e];
	return a->r_cap;
}

inline void OptTransport::SetRCap(Arc* a, double new_rcap)
{
	assert(new_rcap >= 0);
	if (a->r_cap == 0)
	{
		Node* i = a->sister->head;
		if (a->next) a->next->prev = a->prev;
		if (a->prev) a->prev->next = a->next;
		else         i->firstSaturated = a->next;
		a->next = i->firstNonsaturated;
		if (a->next) a->next->prev = a;
		a->prev = NULL;
		i->firstNonsaturated = a;
	}
	a->r_cap = new_rcap;
	if (a->r_cap == 0)
	{
		Node* i = a->sister->head;
		if (a->next) a->next->prev = a->prev;
		if (a->prev) a->prev->next = a->next;
		else         i->firstNonsaturated = a->next;
		a->next = i->firstSaturated;
		if (a->next) a->next->prev = a;
		a->prev = NULL;
		i->firstSaturated = a;
	}
}

inline void OptTransport::SetRCap(EdgeId e, double new_rcap)
{
	SetRCap(&arcs[2*e], new_rcap);
}

inline double OptTransport::GetReverseRCap(EdgeId e)
{
	Arc* a = &arcs[2*e+1];
	return a->r_cap;
}

inline void OptTransport::SetReverseRCap(EdgeId e, double new_rcap)
{
	SetRCap(&arcs[2*e+1], new_rcap);
}

inline void OptTransport::PushFlow(Arc* a, double delta)
{
	if (delta < 0) { a = a->sister; delta = -delta; }
	DecreaseRCap(a, delta);
	IncreaseRCap(a->sister, delta);
	a->head->excess += delta;
	a->sister->head->excess -= delta;
	cost += delta*a->cost;
	if (a->head->excess > 0 && !a->head->next)
	{
		a->head->next = firstActive;
		firstActive = a->head;
	}
}

inline void OptTransport::PushFlow(EdgeId e, double delta)
{
	PushFlow(&arcs[2*e], delta);
}

inline void OptTransport::UpdateCost(EdgeId e, double cap_orig, double delta)
{
	Arc* a = &arcs[2*e];
	cost += delta*(cap_orig-a->r_cap);
	a->cost += delta;
	a->sister->cost = -a->cost;

	if (a->GetRCost() > 0) a = a->sister;
	if (a->r_cap > 0 && a->GetRCost() < 0) PushFlow(a, a->r_cap);
}

///////////////////////////////////////
///////////////////////////////////////
///////////////////////////////////////


void OptTransport::TestOptimality()
{
	Node* i;
	Arc* a;

	for (i=nodes; i<nodes+nodeNum; i++)
	{
		if (fabs(i->excess) > 1e-6)
		{
			assert(0);
		}
		for (a=i->firstSaturated; a; a=a->next)
		{
			if (fabs(a->r_cap) > 1e-6)
			{
				assert(0);
			}
		}
		for (a=i->firstNonsaturated; a; a=a->next)
		{
			double c = a->GetRCost();
			if (a->r_cap <= 0 || a->GetRCost() < -1e-6)
			{
				assert(0);
			}
		}
	}
}

int main_testOT()
{
	const int A = 2, B = 3;
	double x[A+B], V[A*B], z[A*B];
	double* y = x+A;
	double sum;

	int a, b;
	for (a=0; a<A+B; a++)
	{
		x[a] = 1 + (rand() % 3);
	}
	for (a=0; a<A*B; a++)
	{
		V[a] = rand() % 4;
	}

	sum = 0;
	for (a=0; a<A; a++) sum += x[a];
	for (a=0; a<A; a++) x[a] /= sum;

	sum = 0;
	for (b=0; b<B; b++) sum += y[b];
	for (b=0; b<B; b++) y[b] /= sum;

	OptTransport OT;
	OT.Solve(A, B, V, x, y, z);

	printf("x=[ "); for (a=0; a<A; a++) printf("%f ", x[a]); printf("]\n");
	printf("y=[ "); for (b=0; b<B; b++) printf("%f ", y[b]); printf("]\n");

	printf("V,z=\n");
	for (a=0; a<A; a++)
	{
		printf("   ");
		for (b=0; b<B; b++) printf("%f ", V[a+A*b]);
		printf("         ");
		for (b=0; b<B; b++) printf("%f ", z[a+A*b]);
		printf("\n");
	}

	return 0;
}
