#ifndef __OPTIONS_READER__
#define __OPTIONS_READER__

#include <stdio.h>
#include "../LRSolver/block.h"
#include <string.h>

class OptionsReader
{
public:
	OptionsReader(int argc, char* argv[]);
	~OptionsReader();

	void Add(const char* info=NULL);
	void Add(const char* name, double& x, const char* info=NULL);
	void Add(const char* name, int&    x, const char* info=NULL);
	void Add(const char* name, const char*&  x, const char* info=NULL);
	void Add(const char* name, bool&  x, const char* info=NULL);
	void Add(const char* name, int&   x, int choices_num, const char** choices_str, const char* info=NULL);
	void Finish();
	
	void PrintUsage();
	
private:
	
	typedef enum
	{
		NONE,
		DOUBLE,
		INT,
		STRING,
		BOOLEAN,
		ENUM
	} Type;
	
	struct Option
	{
		Option(Type _type, const char* _info, Buffer* buf) : type(_type)
		{
			int len;
			char* ptr;
			if (_info)
			{
				len = (int)strlen(_info)+1; ptr = (char*) buf->Alloc(len); memcpy(ptr, _info, len); info = ptr;
			}
			else info = NULL;
		}
		Option(Type _type, const char* _name, const char* _info, Buffer* buf) : type(_type)
		{
			int len;
			char* ptr;
			len = (int)strlen(_name)+1; ptr = (char*) buf->Alloc(len); memcpy(ptr, _name, len); name = ptr;
			if (_info)
			{
				len = (int)strlen(_info)+1; ptr = (char*) buf->Alloc(len); memcpy(ptr, _info, len); info = ptr;
			}
			else info = NULL;
		}
		Type type;
		const char* name;
		const char* info;
		Option* next;
	};
	
	struct OptionNONE : Option
	{
		OptionNONE(const char* _info, Buffer* buf) : Option(NONE, _info, buf) { }
	};
	struct OptionDOUBLE : Option
	{
		OptionDOUBLE(const char* _name, double& _x, const char* _info, Buffer* buf) : Option(DOUBLE, _name, _info, buf), x(_x) { }
		double& x;
	};
	struct OptionINT : Option
	{
		OptionINT(const char* _name, int& _x, const char* _info, Buffer* buf) : Option(INT, _name, _info, buf), x(_x) { }
		int& x;
	};
	struct OptionSTRING : Option
	{
		OptionSTRING(const char* _name, const char*& _x, const char* _info, Buffer* buf) : Option(STRING, _name, _info, buf), x(_x) { }
		const char*& x;
	};
	struct OptionBOOLEAN : Option
	{
		OptionBOOLEAN(const char* _name, const bool& _x, const char* _info, Buffer* buf) : Option(BOOLEAN, _name, _info, buf), x(_x) { } 
		const bool& x;
	};
	struct OptionENUM : Option
	{
		OptionENUM(const char* _name, int& _x, int _choices_num, const char** _choices_str, const char* _info, Buffer* buf) : Option(ENUM, _name, _info, buf), x(_x)
		{
			choices_num = _choices_num;
			choices_str = (const  char**) buf->Alloc(choices_num*sizeof(const char*));
			int i;
			for (i=0; i<choices_num; i++)
			{
				int len;
				char* ptr;
				len = (int)strlen(_choices_str[i])+1; ptr = (char*) buf->Alloc(len); memcpy(ptr, _choices_str[i], len); choices_str[i] = ptr;
			}
		} 
		int& x;
		int choices_num;
		const char** choices_str;
	};

	char* GetStr(const char* name)
	{
		int name_len, k = -1;
		for (name_len=0; name[name_len]; name_len++) { if (name[name_len] == '#') break; }
		if (name[name_len] == '#')
		{
			k = atoi(name+name_len+1);
		}

		int i, j;
		for (i=0; i<argc; i++)
		{
			char* s = argv[i];
			if (strlen(s) > name_len && s[name_len] == '=' && strncmp(name, s, name_len) == 0)
			{
				char* value = s+name_len+1;
				if (k >= 0)
				{
					for ( ; k>0; k--)
					{
						for (j=0; value[j]; j++) { if (value[j] == ':') break; }
						if (value[j] != ':') { printf("Fatal error in OptionsReader\n"); exit(1); }
						value += j+1;
					}
					for (j=0; value[j]; j++)
					{
						if (value[j] == ':')
						{
							char* value_new = (char*) buffer.Alloc(j+1);
							memcpy(value_new, value, j);
							value_new[j] = 0;
							return value_new;
						}
					}
				}
				for ( ; i<argc-1; i++) argv[i] = argv[i+1];
				argc --;
				return value;
			}
		}
		return NULL;
	}

	Buffer buffer;
	int argc;
	char** argv;
	Option* option_first;
	Option* option_last;

	////////////////////////////

	ReusableBuffer rbuf;
	char* LINE;
	void InitLINE() { LINE = (char*) rbuf.Alloc(16); LINE[0] = 0; }
	void AddToLINE(const char* s)
	{
		if (!s) s = "NULL";
		int a = (int)strlen(LINE);
		int b = (int)strlen(s);
		LINE = (char*) rbuf.Realloc(a+b+1); 
		memcpy(LINE+a, s, b+1);
	}
	void AddToLINE(int x)
	{
		int a = (int)strlen(LINE);
		int b = snprintf(LINE+a, 0, "%d", x);
		LINE = (char*) rbuf.Realloc(a+b+2);
		snprintf(LINE+a, b+1, "%d", x);
	}
	void AddToLINE(double x)
	{
		int a = (int)strlen(LINE);
		int b = snprintf(LINE+a, 0, "%f", x);
		LINE = (char*) rbuf.Realloc(a+b+2);
		snprintf(LINE+a, b+1, "%f", x);
	}
};

inline OptionsReader::OptionsReader(int _argc, char* _argv[]) : argc(_argc), argv(_argv), option_first(NULL), option_last(NULL), buffer(512)
{
}

inline OptionsReader::~OptionsReader()
{
	while (option_first)
	{
		Option* o = option_first;
		option_first = o->next;
		if      (o->type == NONE)    delete ((OptionNONE*)o);
		if      (o->type == DOUBLE)  delete ((OptionDOUBLE*)o);
		else if (o->type == INT)     delete ((OptionINT*)o);
		else if (o->type == STRING)  delete ((OptionSTRING*)o);
		else if (o->type == BOOLEAN) delete ((OptionBOOLEAN*)o);
		else if (o->type == ENUM)    delete ((OptionENUM*)o);
	}
}

inline void OptionsReader::Add(const char* info)
{
	OptionNONE* o = new OptionNONE(info, &buffer);
	if (option_last) option_last->next = o;
	else             option_first = o;
	option_last = o;
	o->next = NULL;
}

inline void OptionsReader::Add(const char* name, double& x, const char* info)
{
	OptionDOUBLE* o = new OptionDOUBLE(name, x, info, &buffer);
	if (option_last) option_last->next = o;
	else             option_first = o;
	option_last = o;
	o->next = NULL;
	
	char* s = GetStr(name);
	if (s) x = atof(s);
}

inline void OptionsReader::Add(const char* name, int& x, const char* info)
{
	OptionINT* o = new OptionINT(name, x, info, &buffer);
	if (option_last) option_last->next = o;
	else             option_first = o;
	option_last = o;
	o->next = NULL;

	char* s = GetStr(name);
	if (s) x = atoi(s);
}

inline void OptionsReader::Add(const char* name, const char*& x, const char* info)
{
	OptionSTRING* o = new OptionSTRING(name, x, info, &buffer);
	if (option_last) option_last->next = o;
	else             option_first = o;
	option_last = o;
	o->next = NULL;

	const char* s = GetStr(name);
	if (s) x = s;
}

inline void OptionsReader::Add(const char* name, bool& x, const char* info)
{
	OptionBOOLEAN* o = new OptionBOOLEAN(name, x, info, &buffer);
	if (option_last) option_last->next = o;
	else             option_first = o;
	option_last = o;
	o->next = NULL;

	int i;
	int name_len = (int)strlen(name);
	for (i=0; i<argc; i++)
	{
		char* s = argv[i];
		if (strlen(s) == name_len+1 && (s[0] == '-' || s[0] == '+') && strncmp(name, s+1, name_len) == 0)
		{
			x = (s[0] == '+') ? true : false;
			for ( ; i<argc-1; i++) argv[i] = argv[i+1];
			argc --;
			break;
		}
	}
}

inline void OptionsReader::Add(const char* name, int& x, int choices_num, const char** choices_str, const char* info)
{
	OptionENUM* o = new OptionENUM(name, x, choices_num, choices_str, info, &buffer);
	if (option_last) option_last->next = o;
	else             option_first = o;
	option_last = o;
	o->next = NULL;

	const char* s = GetStr(name);
	if (s)
	{
		for (x=0; x<choices_num; x++)
		{
			if (strcmp(s, choices_str[x]) == 0) return;
		}
		printf("Error: unknown option (%s=%s)\n", name, s);
		exit(1);
	}
}


inline void OptionsReader::PrintUsage()
{
	Option* o;
	printf("List of options; default values are inside <...> or marked with '*' (e.g. [+|-*] means '+' or '-' with '-' being default):\n");

	int run;
	int len_max = 0;
	for (run=0; run<2; run++)
	{
		for (o=option_first; o; o=o->next)
		{
			InitLINE();
			if (o->type != NONE)
			{
				AddToLINE("    ");

				int r;
				for (r=0; o->name[r]; r++) { if (o->name[r] == '#') break; }
				if (o->name[r] == '#' && o->name[r+1] == '0' && o->name[r+2] == 0)
				{
					const int R = 100;
					char* NAME = new char[r+R];
					memcpy(NAME, o->name, r);
					NAME[r] = 0;
					AddToLINE(NAME);
					AddToLINE("=<");
					int k;
					for ( k=0; ; k++ )
					{
						if (k) AddToLINE(":");
						if      (o->type == INT)     AddToLINE(((OptionINT*)o)->x);
						else if (o->type == DOUBLE)  AddToLINE(((OptionDOUBLE*)o)->x);
						else { printf("Fatal error in OptionsReader\n"); exit(1); }
						if (!o->next || !o->next->name || o->next->type == NONE) break;
						snprintf(NAME+r, R, "#%d", k+1);
						if (strcmp(NAME, o->next->name)) break;
						o = o->next;
					}
					AddToLINE(">");
					delete [] NAME;
				}
				else
				{
					AddToLINE(o->name);
					if      (o->type == BOOLEAN) AddToLINE( (((OptionBOOLEAN*)o)->x) ? "[+*|-]" : "[+|-*]");
					else if (o->type == INT)     { AddToLINE("=<"); AddToLINE(((OptionINT*)o)->x); AddToLINE(">"); }
					else if (o->type == DOUBLE)  { AddToLINE("=<"); AddToLINE(((OptionDOUBLE*)o)->x); AddToLINE(">"); }
					else if (o->type == STRING)  { AddToLINE("=<"); AddToLINE(((OptionSTRING*)o)->x); AddToLINE(">"); }
					else if (o->type == ENUM)
					{
						OptionENUM* oo = (OptionENUM*)o;
						AddToLINE("=[");
						int i;
						for (i=0; i<oo->choices_num; i++)
						{
							AddToLINE(oo->choices_str[i]);
							if (oo->x == i) AddToLINE("*");
							AddToLINE((i==oo->choices_num-1) ? "]" : "|");
						}
					}
				}
			}

			if (o->info && o->info[0] == '\n') continue;
			int len = (int)strlen(LINE);
			if (run == 0) { if (len_max < len) len_max = len; }
			else
			{
				printf("%s", LINE);
				if (o->type == NONE && o->info && o->info[0] == '>')
				{
					printf("  %s", o->info);
				}
				else
				{
					while (len ++ < len_max) printf(" ");
					if (o->info) printf("    %s", o->info);
				}
				printf("\n");
			}
		}
	}
	exit(1);
}

inline void OptionsReader::Finish()
{
	if (argc != 1)
	{
		if (strcmp(argv[1], "help")) printf("Unknown option: %s\n", argv[1]);
		PrintUsage();
	}
}

#endif

