#ifndef __PAIRWISE_TERM_POTTS__
#define __PAIRWISE_TERM_POTTS__

struct PairwiseTermPotts : PairwiseTerm
{
	double get(int a, int b, int A, int B);
	void pass_message(double* Da, double* Db, int* back_ptr_b, int K);
	double pass_message_srmp(double* minus_Da, double* Db, int K);
	void print(int K);

	bool explicit_contraction(int A, int B, int AA, int BB) { return false; }
	void pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B,
		int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift);

	double lambda;
};

inline PairwiseTermPotts* AllocPairwiseTermPotts(int K, double lambda, Buffer* buf)
{
	PairwiseTermPotts* T = (PairwiseTermPotts*) buf->Alloc(sizeof(PairwiseTermPotts));
	T->type = PairwiseTerm::Type::POTTS;

	T->lambda = lambda;
	return T;
}

inline PairwiseTermPotts* AllocPairwiseTermPotts(int A, int B, double* V, Buffer* buf)
{
	if (A != B || A < 2) return NULL;
	int a, b;
	for (a=0; a<A; a++)
	{
		if (V[a + a*A] != 0) return NULL;
	}
	double lambda = V[1];
	if (lambda <= 0) return NULL;
	for (a=0; a<A; a++)
	for (b=0; b<B; b++)
	{
		if (a != b && V[a + A*b] != lambda) return NULL;
	}
	return AllocPairwiseTermPotts(A, lambda, buf);
}

inline double PairwiseTermPotts::get(int a, int b, int A, int B)
{
	return (a == b) ? 0 : lambda;
}

inline void PairwiseTermPotts::pass_message(double* Da, double* Db, int* back_ptr_b, int K)
{
	int a, a_min = 0;
	double Da_min = Da[0];
	for (a=1; a<K; a++)
	{
		if (Da_min > Da[a]) { Da_min = Da[a]; a_min = a; }
	}
	double Da_min_lambda = Da_min + lambda;
	for (a=0; a<K; a++)
	{
		if (a == a_min) { Db[a] += Da_min; back_ptr_b[a] = a_min; }
		else
		{
			if (Da[a] > Da_min_lambda) { Db[a] += Da_min_lambda; back_ptr_b[a] = a_min; }
			else                       { Db[a] += Da[a]; back_ptr_b[a] = a; }
		}
	}
}

inline double PairwiseTermPotts::pass_message_srmp(double* minus_Da, double* Db, int K)
{
	int a, a_min = 0;
	double Da_min = -minus_Da[0];
	for (a=1; a<K; a++)
	{
		if (Da_min > -minus_Da[a]) { Da_min = -minus_Da[a]; a_min = a; }
	}
	for (a=0; a<K; a++)
	{
		if (a == a_min) { Db[a] = 0; }
		else
		{
			double v = -minus_Da[a] - Da_min;
			Db[a] = (v < lambda) ? v : lambda;
		}
	}
	return Da_min;
}

inline void PairwiseTermPotts::print(int K)
{
	printf("[%f]\n", lambda);
}

inline void PairwiseTermPotts::pass_message_fw(double* Da, double* Db, int* back_ptr_b, int A, int B,
	int AA, int BB, int* list_a, int* list_b, int a_shift, int b_shift)
{
	int a, b, a_min = 0;
	double Da_min = Da[0];
	for (a=1; a<AA; a++)
	{
		if (Da_min > Da[a]) { Da_min = Da[a]; a_min = a; }
	}
	double Da_min_lambda = Da_min + lambda;
	int shift = a_shift-b_shift;
	for (a=b=0; b<BB; b++)
	{
		while (list_a[a]-shift < list_b[b])
		{
			a ++;
			if (a == AA)
			{
				for ( ; b<BB; b++)
				{
					Db[b] += Da_min_lambda; back_ptr_b[b] = a_min;
				}
				return;
			}
		}
		if (list_a[a]-shift == list_b[b])
		{
			if (a == a_min || Da[a] <= Da_min_lambda) { Db[b] += Da[a]; back_ptr_b[b] = a; }
			else                                      { Db[b] += Da_min_lambda; back_ptr_b[b] = a_min; }
		}
		else
		{
			Db[b] += Da_min_lambda; back_ptr_b[b] = a_min;
		}
	}
}

#endif
